/*
 * linux/drivers/mtd/maps/armadilloj.c
 *
 * Copyright (C) 2001-2003 Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-J map driver"
#define LICENSE "GPL v2"

static inline void disable_write_protect(void)
{
        unsigned int val = readl(NETARM_MEM_MODULE_BASE + NETARM_MEM_CS0_BASE_ADDR);
        writel(val &= ~(NETARM_MEM_BAR_WRT_PROT), NETARM_MEM_MODULE_BASE + NETARM_MEM_CS0_BASE_ADDR);
}

__u8 armadilloj_read8(struct map_info *map, unsigned long ofs)
{
	DEBUG(2, "armadilloj_read8: 0x%lx -> [0x%x]\n", ofs, readb(map->map_priv_1 + ofs));
	return readb(map->map_priv_1 + ofs);
}

__u16 armadilloj_read16(struct map_info *map, unsigned long ofs)
{
	DEBUG(2, "armadilloj_read16: 0x%lx -> [0x%x]\n", ofs, readw(map->map_priv_1 + ofs));
	return readw(map->map_priv_1 + ofs);
}

void armadilloj_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	DEBUG(2, "armadilloj_write16: [0x%x]@%p -> 0x%lx\n", d, &d, adr);
	writew(d, map->map_priv_1 + (adr));
	mb();
}

void armadilloj_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	DEBUG(2, "armadilloj_copy_from: 0x%x@0x%lx -> 0x%p\n", len, from, to);
	memcpy(to, map->map_priv_1 + from, len);
}

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static struct mtd_info *flash_mtdinfo;

struct map_info armadilloj_flash_map = {
	.name      = "Flash",
	.size      = FLASH_SIZE,
	.buswidth  = FLASH_WIDTH,

	.read8     = armadilloj_read8,
	.read16    = armadilloj_read16,
	.read32    = 0,
	.copy_from = armadilloj_copy_from,

	.write8    = 0,
	.write16   = armadilloj_write16,
    .write32   = 0,
	.copy_to   = 0,
};

/*
 *   0x02000000 +-------------------------+
 *              |      1 Reserved(boot)   | 
 *   0x02020000 +-------------------------+
 *              |         2 Hermit        |
 *   0x02040000 +--------------+----------+
 *              |  3 Image     | 4 Kernel |
 *   0x020B0000 |              +----------+
 *              |              | 5 User   |
 *   0x021E0000 +----------+   |          |
 *              | 6 Backup |   |          |
 *   0x021F0000 +----------+   |          |
 *              | 7 Config |   |          |
 *              +----------+---+----------+
 */

struct mtd_partition armadilloj_partitions[] = {
	{
		.name       = "Flash/Reserved",
		.size       = RESERVED_SIZE,
		.offset     = RESERVED_OFFSET,
		.mask_flags = MTD_WRITEABLE, /* force read-only */
	},
	{
		.name       = "Flash/Hermit",
		.size       = HERMIT_SIZE,
		.offset     = HERMIT_OFFSET,
		.mask_flags = MTD_WRITEABLE, /* force read-only */
	},
	{
		.name       = "Flash/Image",
		.size       = IMAGE_SIZE,
		.offset     = IMAGE_OFFSET,
	},
	{
		.name       = "Flash/Kernel",
		.size       = KERNEL_SIZE,
		.offset     = KERNEL_OFFSET,
	},
	{
		.name       = "Flash/User",
		.size       = USER_SIZE,
		.offset     = USER_OFFSET,
	},
	{
		.name       = "Flash/Backup",
		.size       = BACKUP_SIZE,
		.offset     = BACKUP_OFFSET,
	},
	{
		.name       = "Flash/Config",
		.size       = CONFIG_SIZE,
		.offset     = CONFIG_OFFSET,
	},
};

struct resource armadilloj_flash_resource = {
	.name  = "Flash",
	.start = FLASH_START,
	.end   = FLASH_START + FLASH_SIZE - 1,
	.flags = IORESOURCE_IO | IORESOURCE_BUSY,
};

static int __init init_armadilloj_flash (void)
{
	struct mtd_partition *parts;
	int nb_parts = 0;
	char *part_type;
	int err;
	
	part_type = "static";
	parts = armadilloj_partitions;
	nb_parts = NB_OF(armadilloj_partitions);
	
	if (request_resource (&ioport_resource, &armadilloj_flash_resource)) {
		printk(KERN_NOTICE "Failed to reserve Armadillo-J FLASH space\n");
		err = -EBUSY;
		goto out;
	}
	
	armadilloj_flash_map.map_priv_1 = (unsigned long)ioremap(FLASH_START, FLASH_SIZE);
	if (!armadilloj_flash_map.map_priv_1) {
		printk(KERN_NOTICE "Failed to ioremap Armadillo-J FLASH space\n");
		err = -EIO;
		goto out_resource;
	}

	flash_mtdinfo = do_map_probe("cfi_probe", &armadilloj_flash_map);
	if (!flash_mtdinfo) {
		flash_mtdinfo = do_map_probe("map_rom", &armadilloj_flash_map);
	}
	if (!flash_mtdinfo) {
		printk("FLASH probe failed\n");
		err = -ENXIO;
		goto out_ioremap;
	}

	flash_mtdinfo->module = THIS_MODULE;
	
	if (add_mtd_partitions(flash_mtdinfo, parts, nb_parts)) {
		printk("FLASH partitions addition failed\n");
		err = -ENOMEM;
		goto out_probe;
	}

        /* we have write protect in mtd layer, so don't bother. keep
         * it disabled while kernel is running */
        disable_write_protect();
		
	return 0;

out_probe:
	map_destroy(flash_mtdinfo);
	flash_mtdinfo = 0;
out_ioremap:
	iounmap((void *)armadilloj_flash_map.map_priv_1);
out_resource:
	release_resource (&armadilloj_flash_resource);
out:
	return err;
}

static int __init init_armadilloj_maps(void)
{

       	printk(KERN_INFO "Initializing Armadillo-J MTD mappings\n",
	       FLASH_SIZE, FLASH_START);

	init_armadilloj_flash();
	
	return 0;
}
	

static void __exit cleanup_armadilloj_maps(void)
{
	if (flash_mtdinfo) {
		del_mtd_partitions(flash_mtdinfo);
		map_destroy(flash_mtdinfo);
		iounmap((void *)armadilloj_flash_map.map_priv_1);
		release_resource (&armadilloj_flash_resource);
	}
}

module_init(init_armadilloj_maps);
module_exit(cleanup_armadilloj_maps);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);
MODULE_DESCRIPTION("Armadillo-J map driver");
MODULE_LICENSE("GPL");
